/*
 *  Windows command Line hacking tool for Apple Gmux control
 *  Licensed under GPL  *
 *
 *  Based on the apple_gmux.c driver:
 *  Original copyright:
 *  
 *  Copyright (C) Canonical Ltd. <seth.forshee@canonical.com>
 *  Copyright (C) 2010-2012 Andreas Heider <andreas@meetr.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 */

// WinGmuxCtrl.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"
#include "winring0/OlsApiInit.h"

/**************************************************************************
  NTDLL.dll methods
**************************************************************************/

typedef NTSTATUS (NTAPI *PFNNTDELAYEXECUTION)(BOOLEAN Alertable, PLARGE_INTEGER Interval);
PFNNTDELAYEXECUTION pfnNtDelayExecution = nullptr;

void udelay(const uint64_t microsec)
{
  LARGE_INTEGER li;
  li.QuadPart = microsec * 10;      // 100-ns units
  pfnNtDelayExecution(FALSE, &li);
}


/**************************************************************************
  Mux Control
**************************************************************************/

/*
* gmux port offsets. Many of these are not yet used, but may be in the
* future, and it's useful to have them documented here anyhow.
*/
#define GMUX_PORT_VERSION_MAJOR		0x0004
#define GMUX_PORT_VERSION_MINOR		0x0005
#define GMUX_PORT_VERSION_RELEASE	0x0006
#define GMUX_PORT_SWITCH_DISPLAY	0x0010
#define GMUX_PORT_SWITCH_GET_DISPLAY	0x0011
#define GMUX_PORT_INTERRUPT_ENABLE	0x0014
#define GMUX_PORT_INTERRUPT_STATUS	0x0016
#define GMUX_PORT_SWITCH_DDC		0x0028
#define GMUX_PORT_SWITCH_EXTERNAL	0x0040
#define GMUX_PORT_SWITCH_GET_EXTERNAL	0x0041
#define GMUX_PORT_DISCRETE_POWER	0x0050
#define GMUX_PORT_MAX_BRIGHTNESS	0x0070
#define GMUX_PORT_BRIGHTNESS		0x0074
#define GMUX_PORT_VALUE			0x00c2
#define GMUX_PORT_READ			0x00d0
#define GMUX_PORT_WRITE			0x00d4

#define GMUX_MIN_IO_LEN			(GMUX_PORT_BRIGHTNESS + 4)

#define GMUX_INTERRUPT_ENABLE		0xff
#define GMUX_INTERRUPT_DISABLE		0x00

#define GMUX_INTERRUPT_STATUS_ACTIVE	0
#define GMUX_INTERRUPT_STATUS_DISPLAY	(1 << 0)
#define GMUX_INTERRUPT_STATUS_POWER	(1 << 2)
#define GMUX_INTERRUPT_STATUS_HOTPLUG	(1 << 3)

#define GMUX_BRIGHTNESS_MASK		0x00ffffff
#define GMUX_MAX_BRIGHTNESS		GMUX_BRIGHTNESS_MASK

#define GMUX_PORT_DPM_CHK 0x00
#define GMUX_PORT_DPM_REG1 0xcc
#define GMUX_PORT_DPM_REG2 0xcd

#define DPM_REG1 0xaa
#define DPM_REG2 0x55

#define GMUX_PORT_DPM_RADDR 0x00d0
#define GMUX_PORT_DPM_WADDRSTAT 0x00d4
#define GMUX_PORT_DPM_DAT0 0x00c2
#define GMUX_PORT_DPM_DAT1 0x00c3
#define GMUX_PORT_DPM_DAT2 0x00c4
#define GMUX_PORT_DPM_DAT3 0x00c5

#define GMUX_DPM_CMD_TO 20
#define GMUX_DPM_CMD_DLY 1 /* in ms */

enum vga_switcheroo_state
{
  VGA_SWITCHEROO_OFF,
  VGA_SWITCHEROO_ON,
};

enum vga_switcheroo_client_id
{
  VGA_SWITCHEROO_IGD,
  VGA_SWITCHEROO_DIS,
  VGA_SWITCHEROO_MAX_CLIENTS
};

/**************************************************************************
  Port i/o
**************************************************************************/

static uint8_t inb(uint16_t port)
{
  uint8_t val = ReadIoPortByte(port);
  fprintf(stdout, "Reading port: 0x%x, value: 0x%x\n", port, val);
  fflush(stdout);
  return val;
}


static void outb(uint8_t val, uint16_t port)
{
  fprintf(stdout, "Writing to port: 0x%x, value: 0x%x\n", port, val);
  fflush(stdout);
  WriteIoPortByte(port, val);
}


/**************************************************************************
  apple_gmux_data
**************************************************************************/

struct apple_gmux_data 
{
  uint16_t iostart;
  uint16_t iolen;

  bool indexed;
  bool is_dpm;

  vga_switcheroo_state power_state;

  CRITICAL_SECTION m_cs;
};

/**************************************************************************
  gmux_index_wait_ready
**************************************************************************/

static int gmux_index_wait_ready(struct apple_gmux_data *gmux_data)
{
  int i = 200;

  uint8_t gwr = inb(gmux_data->iostart + (uint16_t)GMUX_PORT_WRITE);

  while (i && (gwr & 0x01)) {
    inb(gmux_data->iostart + (uint16_t)GMUX_PORT_READ);
    gwr = inb(gmux_data->iostart + (uint16_t)GMUX_PORT_WRITE);
    udelay(100);
    i--;
  }

  return !!i;
}

/**************************************************************************
  gmux_is_indexed
**************************************************************************/

static bool gmux_is_indexed(struct apple_gmux_data *gmux_data)
{
  uint16_t val;

  outb(0xaa, gmux_data->iostart + (uint16_t)0xcc);
  outb(0x55, gmux_data->iostart + (uint16_t)0xcd);
  outb(0x00, gmux_data->iostart + (uint16_t)0xce);

  val = (uint16_t)inb(gmux_data->iostart + (uint16_t)0xcc) |
    ((uint16_t)inb(gmux_data->iostart + (uint16_t)0xcd) << 8);

  if (val == 0x55aa)
    return true;

  return false;
}

/**************************************************************************
  gmux_index_wait_complete
**************************************************************************/

static int gmux_index_wait_complete(struct apple_gmux_data *gmux_data)
{
  int i = 200;
  uint8_t gwr = inb(gmux_data->iostart + (uint16_t)GMUX_PORT_WRITE);

  while (i && !(gwr & 0x01)) {
    gwr = inb(gmux_data->iostart + (uint16_t)GMUX_PORT_WRITE);
    udelay(100);
    i--;
  }

  if (gwr & 0x01)
    inb(gmux_data->iostart + (uint16_t)GMUX_PORT_READ);

  return !!i;
}


/**************************************************************************
  gmux_dpm_cmd_rdy
**************************************************************************/

static inline uint8_t gmux_dpm_cmd_rdy(void)
{
  uint8_t status = inb((uint16_t)GMUX_PORT_DPM_WADDRSTAT);
  uint8_t to = GMUX_DPM_CMD_TO;

  while (status == 0 && to)
  {
    inb((uint16_t)GMUX_PORT_DPM_RADDR);
    Sleep(GMUX_DPM_CMD_DLY);
    status = inb((uint16_t)GMUX_PORT_DPM_WADDRSTAT);
    to--;
    Sleep(GMUX_DPM_CMD_DLY);
  }
  return to != 0;
}

/**************************************************************************
  gmux_dpm_cmd_done
**************************************************************************/

static inline uint8_t gmux_dpm_cmd_done(void)
{
  uint8_t status = 0;
  uint8_t to = GMUX_DPM_CMD_TO;

  do {
    status = inb((uint16_t)GMUX_PORT_DPM_WADDRSTAT);
    to--;
    Sleep(GMUX_DPM_CMD_DLY);
  } while (status == 0 && to);

  if (status == 0)
    inb((uint16_t)GMUX_PORT_DPM_RADDR);

  return to != 0;
}

/**************************************************************************
  gmux_dpm_read
**************************************************************************/

static inline int gmux_dpm_read(struct apple_gmux_data *gmux_data, uint8_t reg,
                                uint8_t size, uint32_t *ret)
{
  *ret = 0x0;

  //   if (down_interruptible(&gmux_data->dpm_lock))
  //     return 0;

  if (!gmux_dpm_cmd_rdy()) {
    fprintf(stdout, "gmux_dpm_cmd_rdy failed.\n");
    fflush(stdout);
    goto gmux_dpm_read_failed;
  }
  WriteIoPortByte(gmux_data->iostart + GMUX_PORT_DPM_RADDR, reg);
  if (!gmux_dpm_cmd_done()) {
    fprintf(stdout, "gmux_dpm_cmd_done failed.\n");
    fflush(stdout);
    goto gmux_dpm_read_failed;
  }
  if (size == 1) {
    *ret = (uint32_t)inb(gmux_data->iostart + (uint16_t)GMUX_PORT_DPM_DAT0);
  } else if (size == 2) {
    *ret = (uint32_t)inb(gmux_data->iostart + (uint16_t)GMUX_PORT_DPM_DAT1) << 8;
    *ret |= (uint32_t)inb(gmux_data->iostart + (uint16_t)GMUX_PORT_DPM_DAT0);
  } else if (size == 4) {
    *ret = (uint32_t)inb(gmux_data->iostart + (uint16_t)GMUX_PORT_DPM_DAT3) << 24;
    *ret |= (uint32_t)inb(gmux_data->iostart + (uint16_t)GMUX_PORT_DPM_DAT2) << 16;
    *ret |= (uint32_t)inb(gmux_data->iostart + (uint16_t)GMUX_PORT_DPM_DAT1) << 8;
    *ret |= (uint32_t)inb(gmux_data->iostart + (uint16_t)GMUX_PORT_DPM_DAT0);
  } else
    goto gmux_dpm_read_failed;

  //up(&gmux_data->dpm_lock);
  return 1;

gmux_dpm_read_failed:

  //up(&gmux_data->dpm_lock);
  return 0;
}

static inline int gmux_dpm_write(struct apple_gmux_data *gmux_data, uint8_t reg,
                                 uint8_t size, uint32_t val)
{
  //   if (down_interruptible(&gmux_data->dpm_lock))
  //     return 0;

  if (size == 1) {
    outb(val, gmux_data->iostart + (uint16_t)GMUX_PORT_DPM_DAT0);
  } else if (size == 2) {
    outb(val>>8, gmux_data->iostart + (uint16_t)GMUX_PORT_DPM_DAT1);
    outb(val, gmux_data->iostart + (uint16_t)GMUX_PORT_DPM_DAT0);
  } else if (size == 4) {
    outb(val>>24, gmux_data->iostart + (uint16_t)GMUX_PORT_DPM_DAT3);
    outb(val>>16, gmux_data->iostart + (uint16_t)GMUX_PORT_DPM_DAT2);
    outb(val>>8, gmux_data->iostart + (uint16_t)GMUX_PORT_DPM_DAT1);
    outb(val, gmux_data->iostart + (uint16_t)GMUX_PORT_DPM_DAT0);
  } else
    goto gmux_dpm_write_failed;

  if (!gmux_dpm_cmd_rdy()) {
    fprintf(stdout, "gmux_dpm_cmd_rdy failed.\n");
    fflush(stdout);
    goto gmux_dpm_write_failed;

  }
  outb(reg, gmux_data->iostart + (uint16_t)GMUX_PORT_DPM_WADDRSTAT);
  if (!gmux_dpm_cmd_done()) {
    fprintf(stdout, "gmux_dpm_cmd_done failed.\n");
    fflush(stdout);
    goto gmux_dpm_write_failed;
  }

  //up(&gmux_data->dpm_lock);
  return 1;

gmux_dpm_write_failed:
  //up(&gmux_data->dpm_lock);
  return 0;
}

static void gmux_index_write8(struct apple_gmux_data *gmux_data, int port,
                              uint8_t val)
{
  EnterCriticalSection(&gmux_data->m_cs);
  outb(val, gmux_data->iostart + GMUX_PORT_VALUE);
  gmux_index_wait_ready(gmux_data);
  outb(port & 0xff, gmux_data->iostart + GMUX_PORT_WRITE);
  gmux_index_wait_complete(gmux_data);
  LeaveCriticalSection(&gmux_data->m_cs);
}

static void gmux_pio_write8(struct apple_gmux_data *gmux_data, int port,
                            uint8_t val)
{
  outb(val, gmux_data->iostart + (uint16_t)port);
}

static void gmux_write8(struct apple_gmux_data *gmux_data, int port, uint8_t val)
{
  if (gmux_data->is_dpm)
  {
    gmux_dpm_write(gmux_data, port, 1, val);
  }
  else
  {
    if (gmux_data->indexed)
      gmux_index_write8(gmux_data, port, val);
    else
      gmux_pio_write8(gmux_data, port, val);
  } 
}


static uint8_t gmux_pio_read8(struct apple_gmux_data *gmux_data, int port)
{
  return inb(gmux_data->iostart + (uint16_t)port);
}

static uint8_t gmux_index_read8(struct apple_gmux_data *gmux_data, int port)
{
  uint8_t val;

  EnterCriticalSection(&gmux_data->m_cs);
  gmux_index_wait_ready(gmux_data);
  outb((port & 0xff), gmux_data->iostart + (uint16_t)GMUX_PORT_READ);
  gmux_index_wait_complete(gmux_data);
  val = inb(gmux_data->iostart + (uint16_t)GMUX_PORT_VALUE);
  LeaveCriticalSection(&gmux_data->m_cs);

  return val;
}

static uint8_t gmux_read8(struct apple_gmux_data *gmux_data, int port)
{
  if (gmux_data->is_dpm) {
    uint32_t ret;
    gmux_dpm_read(gmux_data, port, 1, &ret);
    return ret;
  }

  if (gmux_data->indexed)
    return gmux_index_read8(gmux_data, port);
  else
    return gmux_pio_read8(gmux_data, port);
}


static uint32_t gmux_pio_read32(struct apple_gmux_data *gmux_data, int port)
{
  return ReadIoPortDword(gmux_data->iostart + (uint16_t)port);
}


static uint32_t gmux_index_read32(struct apple_gmux_data *gmux_data, int port)
{
  uint32_t val;

  EnterCriticalSection(&gmux_data->m_cs);
  gmux_index_wait_ready(gmux_data);
  outb((port & 0xff), gmux_data->iostart + (uint16_t)GMUX_PORT_READ);
  gmux_index_wait_complete(gmux_data);
  val = inb(gmux_data->iostart + (uint16_t)GMUX_PORT_VALUE);
  LeaveCriticalSection(&gmux_data->m_cs);

  return val;
}


static uint32_t gmux_read32(struct apple_gmux_data *gmux_data, int port)
{
  if (gmux_data->indexed)
    return gmux_index_read32(gmux_data, port);
  else
    return gmux_pio_read32(gmux_data, port);
}


/**************************************************************************
State Control
**************************************************************************/


static int gmux_set_discrete_state(struct apple_gmux_data *gmux_data,
                                   enum vga_switcheroo_state state)
{ 
  if (state == VGA_SWITCHEROO_ON) {
    gmux_write8(gmux_data, GMUX_PORT_DISCRETE_POWER, 1);
    gmux_write8(gmux_data, GMUX_PORT_DISCRETE_POWER, 3);
    fprintf(stdout, "WinGmuxCtrl:  Discrete card is powered ON\n");
    fflush(stdout);
  } else {
    gmux_write8(gmux_data, GMUX_PORT_DISCRETE_POWER, 1);
    gmux_write8(gmux_data, GMUX_PORT_DISCRETE_POWER, 0);
    fprintf(stdout, "WinGmuxCtrl:  Discrete card is powered OFF\n");
    fflush(stdout);
  }

  gmux_data->power_state = state;

  Sleep(1);

  return 0;
}


/**************************************************************************
AppleMuxControl Class
**************************************************************************/

class AppleMuxControl
{
public:

  AppleMuxControl()
  {
    //
    // Initialize WinRing0

    if(!InitWinRing0()) {
      fprintf(stdout, "Initialization of WinRing0 driver failed!\n");
      fflush(stdout);
      exit(-1);
    }

    InitMux();
  }

  ~AppleMuxControl()
  {

  }

  //////////////
  // COMMANDS //
  //////////////

  int SetPowerState( enum vga_switcheroo_client_id id,
    enum vga_switcheroo_state state)
  {
    if (id == VGA_SWITCHEROO_IGD) {
      return 0;
    }

    gmux_set_discrete_state(&m_gmux, state);
    return 0;
  }

  int SwitchTo(enum vga_switcheroo_client_id id)
  {
    if (id == VGA_SWITCHEROO_IGD) {
      gmux_write8(&m_gmux, GMUX_PORT_SWITCH_DDC, 1);
      gmux_write8(&m_gmux, GMUX_PORT_SWITCH_DISPLAY, 2);
      gmux_write8(&m_gmux, GMUX_PORT_SWITCH_EXTERNAL, 2);
    } else {
      gmux_write8(&m_gmux, GMUX_PORT_SWITCH_DDC, 2);
      gmux_write8(&m_gmux, GMUX_PORT_SWITCH_DISPLAY, 3);
      gmux_write8(&m_gmux, GMUX_PORT_SWITCH_EXTERNAL, 3);
    }

    return 0;
  }

private:

  bool InitWinRing0()
  {
    BOOL res = InitOpenLibSys(&m_osllib);

    if(!res) {
      fprintf(stdout, "Error loading WinRing0 driver!\n");
      fflush(stdout);
      return false;
    }

    return TRUE;
  }

  void InitMux()
  {
    InitializeCriticalSection(&m_gmux.m_cs);

    //
    // Currently hardcoded

    m_gmux.iostart = 0x0700;
    m_gmux.iolen = 0x00FE;

    m_gmux.indexed = false;
    m_gmux.is_dpm = false;
    m_gmux.power_state = VGA_SWITCHEROO_ON;

    dpm_chk = gmux_read8(&m_gmux, GMUX_PORT_DPM_CHK);
    ver_major = gmux_read8(&m_gmux, GMUX_PORT_VERSION_MAJOR);
    ver_minor = gmux_read8(&m_gmux, GMUX_PORT_VERSION_MINOR);
    ver_release = gmux_read8(&m_gmux, GMUX_PORT_VERSION_RELEASE);

    printf("DPM chk: %u, ver_major: %u\n", dpm_chk, ver_major);

    if (dpm_chk != ver_major) 
    {
      /*
      * On some machines the gmux is in ACPI even thought the
      * machine doesn't really have a gmux. Check for invalid
      * version information to detect this.
      */
      
      ver_minor = gmux_read8(&m_gmux, GMUX_PORT_VERSION_MINOR);
      ver_release = gmux_read8(&m_gmux, GMUX_PORT_VERSION_RELEASE);
    
      if (ver_major == 0xff && ver_minor == 0xff && ver_release == 0xff) 
      {
          fprintf(stdout, "gmux device seems to be not present\n");
          fflush(stdout);
          exit(-1);
      }
      m_gmux.is_dpm = 0;
    } 
    else 
    { 
      /* possibly a dp micro variant, found in MBP 10,1 */
      /* check presence */
      
      gmux_write8(&m_gmux, GMUX_PORT_DPM_REG1, DPM_REG1);
      gmux_write8(&m_gmux, GMUX_PORT_DPM_REG2, DPM_REG2);
      
      if (gmux_read8(&m_gmux, GMUX_PORT_DPM_REG1) == DPM_REG1 &&
          gmux_read8(&m_gmux,GMUX_PORT_DPM_REG2) == DPM_REG2) 
      {
          uint32_t version = 0;
          m_gmux.is_dpm = 1;

          if (!gmux_dpm_read(&m_gmux, GMUX_PORT_VERSION_MAJOR, 4, &version)) 
          {
              fprintf(stdout, "could not obtain version information from gmux. Bailing out.\n");
              fflush(stdout);
              exit(-1);
          }
          ver_major = (version >> 24) & 0xff;
          ver_minor = (version >> 16) & 0xff;
          ver_release = (version >> 8) & 0xff;
      } else 
      {
        fprintf(stdout, "gmux device seems to be not present\n");
        fflush(stdout);
        exit(-1);
      }
    }

    fprintf(stdout, "gmux version: %u.%u (release: %u), indexed: %u\n", 
      ver_major,
      ver_minor,
      ver_release,
      m_gmux.indexed);
    
    fflush(stdout);
  }

  void DestroyMux()
  {
    DeleteCriticalSection(&m_gmux.m_cs);
  }

  HMODULE m_osllib;          // WinRing0 Handle 
  apple_gmux_data m_gmux;    // GMUX data

  uint8_t ver_major, ver_minor, ver_release, dpm_chk;      // GMUX version data

};


int main(int argc, char **argv)
{
  pfnNtDelayExecution = (PFNNTDELAYEXECUTION) GetProcAddress(
    GetModuleHandle(_T("ntdll.dll")), "NtDelayExecution");

  if(pfnNtDelayExecution == nullptr)
  {
    fprintf(stdout, "Unable to find NtDelayExection call in ntdll.dll\n");
    exit(-1);
  }
  else
  {
    AppleMuxControl mux;

    if(argc != 3) {
      fprintf(stdout, "Error in arguments!\n\n");
      fprintf(stdout, "Usage:\n\n");
      return -1;
    }

    if(strcmp(argv[1], "switch") == 0)
    {
      if(strcmp(argv[2], "igd") == 0)
      {
        fprintf(stdout, "Switching GPU to IGD...\n");
        fflush(stdout);
        mux.SwitchTo(VGA_SWITCHEROO_IGD);
        fprintf(stdout, "Powering down dGPU...\n");
        fflush(stdout);
        mux.SetPowerState(VGA_SWITCHEROO_DIS, VGA_SWITCHEROO_OFF);
        Sleep(500);
      }
      else
        if(strcmp(argv[2], "dis") == 0)
        {
          fprintf(stdout, "Powering up dGPU...\n");
          fflush(stdout);
          mux.SetPowerState(VGA_SWITCHEROO_DIS, VGA_SWITCHEROO_ON);
          Sleep(1000);
          fprintf(stdout, "Switching GPU to dGPU...\n");
          fflush(stdout);
          mux.SwitchTo(VGA_SWITCHEROO_DIS);
          Sleep(500);
        }
        else
        {
          fprintf(stdout, "Unrecognized parameter\n");
          exit(-1);
        }
    }
  }

  return 0;
}

